@@ -6,7 +6,7 @@
 //! with a Prolog-like syntax.
 //!
 //! ```text
-//! === Avalog 0.4 ===
+//! === Avalog 0.5 ===
 //! Type `help` for more information.
 //! > parent'(alice) : mom
 //! > (bob, parent'(alice))
@@ -166,6 +166,8 @@ mod parsing;
 pub enum Expr {
     /// A symbol.
     Sym(Arc<String>),
+    /// A variable.
+    Var(Arc<String>),
     /// A relation between two symbols.
     Rel(Box<Expr>, Box<Expr>),
     /// A 1-avatar of some expression.
@@ -206,7 +208,7 @@ pub enum Expr {
 impl fmt::Display for Expr {
     fn fmt(&self, w: &mut std::fmt::Formatter<'_>) -> std::result::Result<(), std::fmt::Error> {
         match self {
-            Sym(a1) => write!(w, "{}", a1)?,
+            Sym(a1) | Var(a1) => write!(w, "{}", a1)?,
             Rel(a, b) => write!(w, "({}, {})", a, b)?,
             Ava(a, b) => write!(w, "{}'({})", a, b)?,
             Inner(a) => write!(w, ".{}", a)?,
@@ -278,8 +280,22 @@ impl fmt::Display for Expr {
     }
 }
 
+impl From<Arc<String>> for Expr {
+    fn from(val: Arc<String>) -> Expr {
+        if let Some(c) = val.chars().next() {
+            if c.is_uppercase() {return Var(val)};
+        }
+        Sym(val)
+    }
+}
+
 impl From<&'static str> for Expr {
-    fn from(val: &'static str) -> Expr {Sym(Arc::new(val.into()))}
+    fn from(val: &'static str) -> Expr {
+        if let Some(c) = val.chars().next() {
+            if c.is_uppercase() {return Var(Arc::new(val.into()))};
+        }
+        Sym(Arc::new(val.into()))
+    }
 }
 
 /// Constructs a unique directive expression, e.g. `uniq a`.
@@ -384,7 +400,7 @@ impl Expr {
                 let new_arg: Vec<Expr> = arg.iter().map(|a| a.eval_lift(eval, true)).collect();
                 Rule(Box::new(new_res), new_arg)
             }
-            Sym(_) => self.clone(),
+            Sym(_) | Var(_) => self.clone(),
             UniqAva(_) => self.clone(),
             Ambiguity(_) => self.clone(),
             Tail => self.clone(),
@@ -414,13 +430,8 @@ impl Expr {
     /// Returns `true` if expression contains no variables.
     pub fn is_const(&self) -> bool {
         match self {
-            Sym(ref a) => {
-                if let Some(c) = a.chars().next() {
-                    !c.is_uppercase()
-                } else {
-                    true
-                }
-            }
+            Var(_) => false,
+            Sym(_) => true,
             App(ref a, ref b) => a.is_const() && b.is_const(),
             Ava(ref a, ref b) => a.is_const() && b.is_const(),
             _ => false
@@ -486,43 +497,28 @@ pub fn bind(e: &Expr, a: &Expr, vs: &mut Vec<(Arc<String>, Expr)>, tail: &mut Ve
             bind(a1, a2, vs, tail) &&
             bind(b1, b2, vs, tail)
         }
-        (&Sym(ref a1), &Sym(ref a2)) => {
-            if let Some(c) = a1.chars().next() {
-                if c.is_uppercase() {
-                    // Look for previous occurences of bound variable.
-                    let mut found = false;
-                    for &(ref b, ref b_expr) in &*vs {
-                        if b == a1 {
-                            if let Some(true) = equal(b_expr, a) {
-                                found = true;
-                                break;
-                            } else {
-                                return false;
-                            }
-                        }
+        (&Sym(ref a1), &Sym(ref a2)) => a1 == a2,
+        (&Var(ref a1), &Sym(_)) => {
+            // Look for previous occurences of bound variable.
+            let mut found = false;
+            for &(ref b, ref b_expr) in &*vs {
+                if b == a1 {
+                    if let Some(true) = equal(b_expr, a) {
+                        found = true;
+                        break;
+                    } else {
+                        return false;
                     }
-                    if !found {
-                        vs.push((a1.clone(), a.clone()));
-                    }
-                    true
-                } else {
-                    a1 == a2
                 }
-            } else {
-                false
             }
+            if !found {
+                vs.push((a1.clone(), a.clone()));
+            }
+            true
         }
-        (&Sym(ref a1), _) if a.is_const() => {
-            if let Some(c) = a1.chars().next() {
-                if c.is_uppercase() {
-                    vs.push((a1.clone(), a.clone()));
-                    true
-                } else {
-                    false
-                }
-            } else {
-                false
-            }
+        (&Var(ref a1), _) if a.is_const() => {
+            vs.push((a1.clone(), a.clone()));
+            true
         }
         (&Ava(ref a1, ref b1), &Ava(ref a2, ref b2)) => {
             bind(a1, a2, vs, tail) &&
@@ -537,18 +533,15 @@ fn substitute(r: &Expr, vs: &Vec<(Arc<String>, Expr)>) -> Expr {
         Rel(a1, b1) => {
             rel(substitute(a1, vs), substitute(b1, vs))
         }
-        Sym(a1) => {
-            if let Some(c) = a1.chars().next() {
-                if c.is_uppercase() {
-                    for v in vs {
-                        if &v.0 == a1 {
-                            return v.1.clone();
-                        }
-                    }
+        Var(a1) => {
+            for v in vs {
+                if &v.0 == a1 {
+                    return v.1.clone();
                 }
             }
             r.clone()
         }
+        Sym(_) => r.clone(),
         Ava(a1, b1) => {
             ava(substitute(a1, vs), substitute(b1, vs))
         }
@@ -566,19 +559,15 @@ fn substitute(r: &Expr, vs: &Vec<(Arc<String>, Expr)>) -> Expr {
         }
         App(a, b) => {
             let a_expr = substitute(a, vs);
-            if let Sym(a1) = &**b {
-                if let Some(c) = a1.chars().next() {
-                    if c.is_uppercase() {
-                        for v in vs {
-                            if &v.0 == a1 {
-                                if let List(args) = &v.1 {
-                                    let mut expr = a_expr;
-                                    for arg in args {
-                                        expr = app(expr, arg.clone());
-                                    }
-                                    return expr;
-                                }
+            if let Var(a1) = &**b {
+                for v in vs {
+                    if &v.0 == a1 {
+                        if let List(args) = &v.1 {
+                            let mut expr = a_expr;
+                            for arg in args {
+                                expr = app(expr, arg.clone());
                             }
+                            return expr;
                         }
                     }
                 }
@@ -607,10 +596,16 @@ fn equal(a: &Expr, b: &Expr) -> Option<bool> {
     else {
         match (a, b) {
             (&Sym(_), &Sym(_)) |
+            (&Sym(_), &Var(_)) |
+            (&Var(_), &Sym(_)) |
+            (&Var(_), &Var(_)) |
             (&App(_, _), &Sym(_)) |
+            (&App(_, _), &Var(_)) |
             (&App(_, _), &Ava(_, _)) => None,
             (&Sym(_), &Ava(_, _)) |
-            (&Ava(_, _), &Sym(_)) => Some(false),
+            (&Var(_), &Ava(_, _)) |
+            (&Ava(_, _), &Sym(_)) |
+            (&Ava(_, _), &Var(_)) => Some(false),
             (&Ava(ref a1, ref b1), &Ava(ref a2, ref b2)) => {
                 let cmp_a = equal(a1, a2);
                 if false_or_none(cmp_a) {return cmp_a};
@@ -723,12 +718,14 @@ pub struct Accelerator {
     pub index: HashMap<Expr, Vec<usize>>,
     /// The number of facts that has been indexed.
     pub len: usize,
+    /// The index of the last rule.
+    pub last_rule: Option<usize>,
 }
 
 impl Accelerator {
     /// Creates a new accelerator.
     pub fn new() -> Accelerator {
-        Accelerator {index: HashMap::new(), len: 0}
+        Accelerator {index: HashMap::new(), len: 0, last_rule: None}
     }
 
     /// Returns a constructor for solve-and-reduce.
@@ -751,7 +748,7 @@ impl Accelerator {
                     insert(i, a);
                     insert(i, b);
                 }
-                Sym(_) | Inner(_) | UniqAva(_) => {
+                Sym(_) | Var(_) | Inner(_) | UniqAva(_) => {
                     insert(i, e);
                 }
                 AmbiguousRel(_, _, _) |
@@ -903,53 +900,39 @@ pub fn infer(solver: Solver<Expr, Accelerator>, facts: &[Expr]) -> Option<Expr>
         }
     }
 
-    // Find the last used rule.
-    // This is used for diagonalization, which means that
-    // rules are expanded using breath-first search.
-    let mut last_rule: Option<(usize, usize)> = None;
-    'outer: for (j, e2) in facts.iter().enumerate().rev() {
-        for (i, e) in facts.iter().enumerate().rev() {
-            if let Rule(_, _) = e {
-                if let Some(new_expr) = match_rule(e, e2) {
-                    if !solver.can_add(&new_expr) {
-                        last_rule = Some((i, j));
-                        break 'outer;
-                    }
-                }
-            }
-        }
-    }
-
-    match last_rule {
+    match solver.accelerator.last_rule {
         None => {
             // Normal rule order.
-            for e in facts {
+            for (i, e) in facts.iter().enumerate() {
                 if let Rule(_, _) = e {
                     for e2 in facts {
                         if let Some(new_expr) = match_rule(e, e2) {
+                            solver.accelerator.last_rule = Some(i);
                             if solver.can_add(&new_expr) {return Some(new_expr)};
                         }
                     }
                 }
             }
         }
-        Some((i, _)) => {
+        Some(i) => {
             // Diagonalize rules.
             // Start with the next rule.
-            for e in &facts[i + 1..] {
+            for (j, e) in facts[i + 1..].iter().enumerate() {
                 if let Rule(_, _) = e {
                     for e2 in facts {
                         if let Some(new_expr) = match_rule(e, e2) {
+                            solver.accelerator.last_rule = Some(i + 1 + j);
                             if solver.can_add(&new_expr) {return Some(new_expr)};
                         }
                     }
                 }
             }
             // Try previous used rules.
-            for e in &facts[..i + 1] {
+            for (j, e) in facts[..i + 1].iter().enumerate() {
                 if let Rule(_, _) = e {
                     for e2 in facts {
                         if let Some(new_expr) = match_rule(e, e2) {
+                            solver.accelerator.last_rule = Some(j);
                             if solver.can_add(&new_expr) {return Some(new_expr)};
                         }
                     }
